# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
Functional tests for prepare-sources command
"""
from pathlib import Path

from packit.utils.commands import cwd
from tests.functional.spellbook import call_real_packit


def test_prepare_sources_command_for_path(upstream_or_distgit_path, tmp_path):
    with cwd(tmp_path):
        call_real_packit(
            parameters=[
                "--debug",
                "prepare-sources",
                "--result-dir",
                Path.cwd(),
                str(upstream_or_distgit_path),
            ],
        )

        tarball_path = next(Path.cwd().glob("*.tar.gz"))
        assert tarball_path.exists()
        specfile_path = next(Path.cwd().glob("*.spec"))
        assert specfile_path.exists()


def test_prepare_sources_command(cwd_upstream_or_distgit):
    call_real_packit(
        parameters=["--debug", "prepare-sources"],
        cwd=cwd_upstream_or_distgit,
    )
    result_dir = cwd_upstream_or_distgit.joinpath("prepare_sources_result")
    assert result_dir.exists()

    tarball_path = next(result_dir.glob("*.tar.gz"))
    assert tarball_path.exists()
    specfile_path = next(result_dir.glob("*.spec"))
    assert specfile_path.exists()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging
import os

logger = logging.getLogger(__name__)


class Pushgateway:
    def __init__(self) -> None:
        try:
            from prometheus_client import CollectorRegistry, Counter
        except ImportError:
            return
        else:
            self.pushgateway_address = os.getenv("PUSHGATEWAY_ADDRESS")
            self.worker_name = os.getenv("HOSTNAME")
            self.registry = CollectorRegistry()

            self.aliases_fallback_used = Counter(
                "aliases_fallback_used",
                "Number of times fallback values for aliases were used",
                registry=self.registry,
            )

    def push(self) -> None:
        try:
            from prometheus_client import push_to_gateway
        except ImportError:
            return
        else:
            if (
                not hasattr(self, "pushgateway_address")
                or not hasattr(self, "worker_name")
                or not hasattr(self, "registry")
            ):
                return
            if not (self.pushgateway_address and self.worker_name):
                logger.debug("Pushgateway address or worker name not defined.")
                return
            logger.info("Pushing the metrics to pushgateway.")
            push_to_gateway(
                self.pushgateway_address,
                job=self.worker_name,
                registry=self.registry,
            )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest
from flexmock import flexmock

import packit
from packit.copr_helper import CoprHelper


class TestCoprHelper:
    @pytest.mark.parametrize(
        # copr_client.mock_chroot_proxy.get_list() returns dictionary
        "get_list_keys, expected_return",
        [
            pytest.param(["chroot1", "_chroot2"], ["chroot1"], id="chroot_list"),
            pytest.param([], [], id="empty_list"),
        ],
    )
    def test_get_avilable_chroots(self, get_list_keys, expected_return):
        copr_client_mock = flexmock(mock_chroot_proxy=flexmock())
        copr_client_mock.mock_chroot_proxy.should_receive("get_list.keys").and_return(
            get_list_keys,
        )
        flexmock(packit.copr_helper.CoprClient).should_receive(
            "create_from_config_file",
        ).and_return(copr_client_mock)

        copr_helper = CoprHelper("_upstream_local_project")
        copr_helper.get_available_chroots.cache_clear()

        assert copr_helper.get_available_chroots() == expected_return

    @pytest.mark.parametrize(
        "owner,project,section,expected_suffix",
        [
            (
                "@rhinstaller",
                "Anaconda",
                "permissions",
                "g/rhinstaller/Anaconda/permissions/",
            ),
            ("@rhinstaller", "Anaconda", None, "g/rhinstaller/Anaconda/edit/"),
            ("someone", "Anaconda", "permissions", "someone/Anaconda/permissions/"),
        ],
    )
    def test_settings_url(self, owner, project, section, expected_suffix):
        copr_client_mock = flexmock(config={"copr_url": "https://fedoracloud.org"})

        flexmock(packit.copr_helper.CoprClient).should_receive(
            "create_from_config_file",
        ).and_return(copr_client_mock)
        copr_helper = CoprHelper("_upstream_local_project")

        assert (
            copr_helper.get_copr_settings_url(owner, project, section)
            == f"https://fedoracloud.org/coprs/{expected_suffix}"
        )

    @pytest.mark.parametrize(
        "update_dict,expect_call_args",
        [
            (
                {},
                None,
            ),
            (
                {"fedora-rawhide-x86_64": {"additional_repos": ([], ["y"])}},
                {
                    "additional_repos": ["y"],
                },
            ),
        ],
    )
    def test_update_chroot_specific_configuration(self, update_dict, expect_call_args):
        project_proxy_mock = flexmock()
        copr_client_mock = flexmock(
            config={"copr_url": "https://fedoracloud.org"},
            project_chroot_proxy=project_proxy_mock,
        )

        flexmock(packit.copr_helper.CoprClient).should_receive(
            "create_from_config_file",
        ).and_return(copr_client_mock)

        if expect_call_args:
            project_proxy_mock.should_receive("edit").with_args(
                projectname="project",
                ownername="owner",
                chrootname="fedora-rawhide-x86_64",
                **expect_call_args,
            ).once()

        copr_helper = CoprHelper("_upstream_local_project")
        copr_helper._update_chroot_specific_configuration(
            project="project",
            owner="owner",
            update_dict=update_dict,
        )

    @pytest.mark.parametrize(
        "targets_dict,result_dict",
        [
            ({"fedora-rawhide": {}}, {}),
            ({"fedora-rawhide": {"distros": ["y"]}}, {}),
            (
                {"fedora-rawhide": {"additional_repos": ["y"]}},
                {
                    "fedora-rawhide-x86_64": {"additional_repos": ([], ["y"])},
                },
            ),
            (
                {
                    "fedora-rawhide": {
                        "additional_modules": "httpd:2.4,nodejs:12",
                        "distros": ["z"],
                    },
                },
                {
                    "fedora-rawhide-x86_64": {
                        "additional_modules": ("", "httpd:2.4,nodejs:12"),
                    },
                },
            ),
        ],
    )
    def test_get_chroot_specific_configuration_to_update(
        self,
        targets_dict,
        result_dict,
    ):
        project_proxy_mock = flexmock()
        copr_client_mock = flexmock(
            config={"copr_url": "https://fedoracloud.org"},
            project_chroot_proxy=project_proxy_mock,
        )

        flexmock(packit.copr_helper.CoprClient).should_receive(
            "create_from_config_file",
        ).and_return(copr_client_mock)

        if result_dict:
            project_proxy_mock.should_receive("get").and_return(
                {
                    "additional_modules": "",
                    "additional_packages": [],
                    "additional_repos": [],
                    "comps_name": None,
                    "delete_after_days": None,
                    "isolation": "unchanged",
                    "mock_chroot": "centos-stream-8-x86_64",
                    "ownername": "@theforeman",
                    "projectname": "pr-testing-playground",
                    "with_opts": [],
                    "without_opts": [],
                },
            )

        copr_helper = CoprHelper("_upstream_local_project")
        assert (
            copr_helper._get_chroot_specific_configuration_to_update(
                "project",
                "owner",
                targets_dict=targets_dict,
            )
            == result_dict
        )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging

import pytest

from packit.config.config import Config

logger = logging.getLogger(__name__)

dep_keys = [
    "github_app_id",
    "github_app_cert_path",
    "github_token",
    "pagure_user_token",
    "pagure_instance_url",
    "pagure_fork_token",
]

authentication = {
    "authentication": {
        "github.com": {"token": "abcd"},
        "pagure": {"token": "abcd", "instance_url": "https://src.fedoraproject.org"},
    },
}


@pytest.mark.parametrize(
    "key",
    dep_keys,
)
def test_with_deprecated_keys(key, caplog, recwarn):
    raw_dict = {key: "somevalue"}
    Config.load_authentication(raw_dict)
    assert "Please, use 'authentication' key in the user" in caplog.text
    caplog.clear()


def test_with_only_authentication(caplog, recwarn):
    Config.load_authentication(authentication)
    assert "Please, use 'authentication' key in the user" not in caplog.text


def test_with_both_authentication(caplog, recwarn):
    raw_dict = authentication
    raw_dict[dep_keys[3]] = "somevalue"
    Config.load_authentication(raw_dict)
    assert "Please, use 'authentication' key in the user" not in caplog.text



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT
import json
from pathlib import PosixPath

import pytest
from click.testing import CliRunner
from flexmock import flexmock

from packit.api import PackitAPI
from packit.cli.builds import koji_build
from packit.cli.packit_base import packit_base
from packit.config import (
    package_config,
)
from packit.distgit import DistGit
from packit.local_project import LocalProject

DEFAULT_CONFIG_YAML = """
    {
        "specfile_path": "python-teamcity-messages.spec",
        "files_to_sync": ["python-teamcity-messages.spec", ".packit.yaml"],
        "upstream_package_name": "teamcity-messages",
        "downstream_package_name": "python-teamcity-messages",
        "upstream_project_url": "https://github.com/majamassarini/teamcity-messages",
        "upstream_tag_template": "v{version}",
        "jobs": [
            {
                "job": "copr_build",
                "trigger": "commit",
                "targets": ["fedora-rawhide", "fedora-stable"]
            },
            {
                "job": "propose_downstream",
                "trigger": "release",
                "dist_git_branches": ["fedora-rawhide", "fedora-stable"]
            },
            {
                "job": "koji_build",
                "trigger": "commit",
                "allowed_pr_authors": ["packit"],
                "dist_git_branches": ["fedora-rawhide", "fedora-stable"]
            },
            {
                "job": "bodhi_update",
                "trigger": "commit",
                "dist_git_branches": ["fedora-rawhide", "fedora-stable"]
            }
        ]
    }
    """


MONOREPO_COPR_PACKIT_YAML = """
    {
        "upstream_project_url": "https://github.com/fedora-copr/copr.git",
        "packages": {
            "python": {
                "downstream_package_name": "python-copr",
                "upstream_package_name": "copr",
                "paths": ["./python"],
                "specfile_path": "python-copr.spec",
                "files_to_sync": ["python-copr.spec"]
            },
            "cli": {
                "downstream_package_name": "copr-cli",
                "upstream_package_name": "copr-cli",
                "paths": ["./cli"],
                "specfile_path": "copr-cli.spec",
                "files_to_sync": ["copr-cli.spec"]
            },
            "frontend": {
                "downstream_package_name": "copr-frontend",
                "upstream_package_name": "copr-frontend",
                "paths": ["./frontend"],
                "specfile_path": "copr-frontend.spec",
                "files_to_sync": ["copr-frontend.spec"]
            }
        },
        "jobs": [
            {
                "job": "copr_build",
                "packages": ["cli", "python"],
                "trigger": "pull_request",
                "targets": "fedora-37",
                "owner": "mmassari",
                "project": "knx-stack"
            },
            {
                "job": "vm_image_build",
                "packages": ["cli", "python"],
                "trigger": "pull_request",
                "copr_chroot": "fedora-37-x86_64",
                "owner": "mmassari",
                "project": "knx-stack",
                "image_customizations": {
                    "packages": ["python-copr", "copr-cli"],
                    "image_distribution": "fedora-37",
                    "image_request": {
                        "architecture": "x86_64",
                        "image_type": "aws",
                        "upload_request": {
                            "type": "aws",
                            "options": {}
                        }
                    }
                }
            }
        ]
    }
    """


@pytest.mark.parametrize(
    "package_config_yaml,mock_api_calls,how_many_times,options",
    [
        pytest.param(
            DEFAULT_CONFIG_YAML,
            [("run_copr_build", ("build_id", "repo_url")), ("watch_copr_build", None)],
            1,
            ["build", "in-copr", "."],
            id="default package config copr build",
        ),
        pytest.param(
            MONOREPO_COPR_PACKIT_YAML,
            [("run_copr_build", ("build_id", "repo_url")), ("watch_copr_build", None)],
            2,
            ["build", "in-copr", "--package=python", "--package=cli"],
            id="monorepo build in copr for python and cli copr packages",
        ),
        pytest.param(
            MONOREPO_COPR_PACKIT_YAML,
            [("run_copr_build", ("build_id", "repo_url")), ("watch_copr_build", None)],
            0,
            ["build", "in-copr", "--package=python", "--package=unknown"],
            id=(
                "monorepo build in copr fails before any action is taken "
                "if a package does not exist"
            ),
        ),
        pytest.param(
            MONOREPO_COPR_PACKIT_YAML,
            [("run_copr_build", ("build_id", "repo_url")), ("watch_copr_build", None)],
            3,
            ["build", "in-copr", "."],
            id="monorepo copr build for all copr packages",
        ),
        pytest.param(
            DEFAULT_CONFIG_YAML,
            [("submit_vm_image_build", "build_id")],
            0,  # there is no image builder job definition
            ["build", "in-image-builder", "--no-wait", "."],
            id="default package config image build",
        ),
        pytest.param(
            MONOREPO_COPR_PACKIT_YAML,
            [("submit_vm_image_build", "build_id")],
            2,
            [
                "build",
                "in-image-builder",
                "--no-wait",
                "--package=python",
                "--package=cli",
                ".",
            ],
            id="monorepo build in image builder for python and cli copr packages",
        ),
        pytest.param(
            MONOREPO_COPR_PACKIT_YAML,
            [("submit_vm_image_build", "build_id")],
            2,  # there is no image builder job definition for frontend package (will fail)
            ["build", "in-image-builder", "--no-wait", "."],
            id="monorepo build in image builder for all copr packages",
        ),
        pytest.param(
            DEFAULT_CONFIG_YAML,
            [("create_srpm", "srpm_path")],
            1,
            ["srpm", "."],
            id="default package config srpm build",
        ),
        pytest.param(
            MONOREPO_COPR_PACKIT_YAML,
            [("create_srpm", "srpm_path")],
            2,
            ["srpm", "--package=cli", "--package=frontend"],
            id="monorepo srpm build for frontend and cli copr packages",
        ),
        pytest.param(
            MONOREPO_COPR_PACKIT_YAML,
            [("create_srpm", "srpm_path")],
            3,
            ["srpm", "."],
            id="monorepo srpm build for all copr packages",
        ),
    ],
)
def test_iterate_packages(package_config_yaml, mock_api_calls, how_many_times, options):
    package_config_dict = json.loads(package_config_yaml)

    flexmock(package_config).should_receive("find_packit_yaml").and_return(
        flexmock(name=".packit.yaml", parent="/some/dir/"),
    )
    flexmock(package_config).should_receive("load_packit_yaml").and_return(
        package_config_dict,
    )
    flexmock(LocalProject).should_receive("git_repo").and_return(
        flexmock(
            remotes=[],
            active_branch=flexmock(name="an active branch"),
            head=flexmock(is_detached=False),
        ),
    )
    # otherwise _dg and _local_project objects will be created
    # by debugger threads and you are not able to debug them
    flexmock(PackitAPI).should_receive("__repr__").and_return("")
    flexmock(DistGit).should_receive("__repr__").and_return("")
    flexmock(koji_build).should_receive("get_branches").and_return("rawhide")
    flexmock(PackitAPI).should_receive("init_kerberos_ticket").and_return()
    for call, result in mock_api_calls:
        flexmock(PackitAPI).should_receive(call).and_return(result).times(
            how_many_times,
        )

    runner = CliRunner()
    runner.invoke(packit_base, options)


@pytest.mark.parametrize(
    "package_config_yaml,mock_api_calls,how_many_times,options,dist_git,dist_git_is_git_repo",
    [
        pytest.param(
            DEFAULT_CONFIG_YAML,
            [("sync_status_string", None)],
            1,
            [
                "source-git",
                "status",
                ".",
                ".",
            ],  # fake dist-git dir for click safety checks
            "python-teamcity-messages",  # mocked dist-git value
            True,
            id="source git status for a default config",
        ),
        pytest.param(
            DEFAULT_CONFIG_YAML,
            [("sync_status_string", None)],
            1,
            [
                "source-git",
                "status",
                ".",
                ".",
            ],  # fake dist-git dir for click safety checks
            "teamcity-messages",  # mocked dist-git value
            True,
            id="source git status for a default config with no matching dist-git repo name",
        ),
        pytest.param(
            DEFAULT_CONFIG_YAML,
            [("sync_status_string", None)],
            0,
            [
                "source-git",
                "status",
                ".",
                ".",
            ],  # fake dist-git dir for click safety checks
            "teamcity-messages",  # mocked dist-git value
            False,
            id="source git status for a default config with no matching dist-git dir",
        ),
        pytest.param(
            MONOREPO_COPR_PACKIT_YAML,
            [("sync_status_string", None)],
            1,
            [
                "source-git",
                "status",
                ".",
                ".",
            ],  # fake dist-git dir for click safety checks
            "copr-cli",  # mocked dist-git value
            True,
            id="source git status for monorepo copr package copr-cli",
        ),
        pytest.param(
            MONOREPO_COPR_PACKIT_YAML,
            [("sync_status_string", None)],
            1,
            [
                "source-git",
                "status",
                ".",
                ".",
            ],  # fake dist-git dir for click safety checks
            "rpms",  # mocked dist-git value
            False,
            id="source git status for all the monorepo copr packages",
        ),
    ],
)
def test_iterate_packages_source_git(
    package_config_yaml,
    mock_api_calls,
    how_many_times,
    options,
    dist_git,
    dist_git_is_git_repo,
):
    package_config_dict = json.loads(package_config_yaml)
    flexmock(PosixPath).should_receive("name").and_return(dist_git)
    flexmock(PosixPath).should_call("joinpath").with_args(str)
    flexmock(PosixPath).should_receive("joinpath").with_args(".git").and_return(
        flexmock().should_receive("exists").and_return(dist_git_is_git_repo).mock(),
    )
    flexmock(PosixPath).should_receive("glob").and_return(
        [
            flexmock(is_dir=lambda: True, name="copr-cli")
            .should_receive("joinpath")
            .and_return(flexmock().should_receive("exists").and_return(True).mock())
            .mock(),
        ],
    )

    flexmock(package_config).should_receive("load_packit_yaml").and_return(
        package_config_dict,
    )
    # otherwise _dg and _local_project objects will be created
    # by debugger threads and you are not able to debug them
    flexmock(PackitAPI).should_receive("__repr__").and_return("")
    for call, result in mock_api_calls:
        flexmock(PackitAPI).should_receive(call).and_return(result).times(
            how_many_times,
        )

    runner = CliRunner()
    runner.invoke(packit_base, options)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import json
import logging
import os
import sys

import click

from packit.cli.types import LocalProjectParameter
from packit.cli.utils import cover_packit_exception, get_packit_api, iterate_packages
from packit.config import (
    get_context_settings,
    pass_config,
)
from packit.constants import (
    PACKAGE_LONG_OPTION,
    PACKAGE_OPTION_HELP,
    PACKAGE_SHORT_OPTION,
)

logger = logging.getLogger(__name__)


@click.command("scan-in-osh", context_settings=get_context_settings())
@pass_config
@cover_packit_exception
@iterate_packages
@click.option(
    PACKAGE_SHORT_OPTION,
    PACKAGE_LONG_OPTION,
    multiple=True,
    help=PACKAGE_OPTION_HELP.format(action="build"),
)
@click.option(
    "--target",
    help="Chroot to build in. (defaults to 'fedora-rawhide-x86_64')",
    default="fedora-rawhide-x86_64",
)
@click.option(
    "--base-srpm",
    help="Base SRPM to perform a differential build against",
    default=None,
)
@click.option(
    "--comment",
    help="Comment for the build",
    default="Submitted through Packit.",
)
@click.option(
    "--csmock-args",
    help="Pass additional arguments to csmock",
    default=None,
)
@click.argument("path_or_url", type=LocalProjectParameter(), default=os.path.curdir)
def scan_in_osh(
    config,
    path_or_url,
    package_config,
    target,
    base_srpm,
    comment,
    csmock_args,
):
    """
    Perform a scan through OpenScanHub.
    You need a valid Kerberos ticket and set `dns_canonicalize_hostname=false`
    in Kerberos configurations.
    Documentation can be found at https://fedoraproject.org/wiki/OpenScanHub.
    """
    api = get_packit_api(
        config=config,
        package_config=package_config,
        local_project=path_or_url,
    )

    if base_srpm:
        logger.debug(f"Base SRPM: {base_srpm}")

    cmd_result_stdout = api.run_osh_build(
        chroot=target,
        base_srpm=base_srpm,
        comment=comment,
        csmock_args=csmock_args,
    )

    if cmd_result_stdout:
        build_url = json.loads(cmd_result_stdout)["url"]
        logger.info(f"Scan URL: {build_url}")
        sys.exit(0)

    sys.exit(1)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import json
import logging
import os
import warnings
from enum import Enum
from functools import lru_cache, partial
from pathlib import Path
from shutil import which
from typing import Optional

import click
from lazy_object_proxy import Proxy
from ogr import GithubService, PagureService, get_instances_from_dict, get_project
from ogr.abstract import GitProject, GitService
from ogr.exceptions import OgrException
from yaml import safe_load

from packit.constants import (
    CONFIG_FILE_NAMES,
    SANDCASTLE_DEFAULT_PROJECT,
    SANDCASTLE_IMAGE,
    SANDCASTLE_PVC,
    SANDCASTLE_WORK_DIR,
)
from packit.exceptions import PackitConfigException, PackitException

logger = logging.getLogger(__name__)


class Config:
    def __init__(
        self,
        debug: bool = False,
        fas_user: Optional[str] = None,
        fas_password: Optional[str] = None,
        keytab_path: Optional[str] = None,
        redhat_api_refresh_token: Optional[str] = None,
        upstream_git_remote: Optional[str] = None,
        kerberos_realm: Optional[str] = "FEDORAPROJECT.ORG",
        koji_build_command: Optional[str] = "koji build",
        pkg_tool: str = "fedpkg",
        command_handler: Optional[str] = None,
        command_handler_work_dir: str = SANDCASTLE_WORK_DIR,
        command_handler_pvc_env_var: str = SANDCASTLE_PVC,
        command_handler_image_reference: str = SANDCASTLE_IMAGE,
        command_handler_k8s_namespace: str = SANDCASTLE_DEFAULT_PROJECT,
        command_handler_pvc_volume_specs: Optional[list[dict[str, str]]] = None,
        command_handler_storage_class: str = "",
        appcode: str = "",
        package_config_path=None,
        repository_cache=None,
        add_repositories_to_repository_cache=True,
        default_parse_time_macros: Optional[dict] = None,
        **kwargs,
    ):
        self.debug: bool = debug
        self.fas_user: Optional[str] = fas_user
        self.fas_password: Optional[str] = fas_password
        self.keytab_path: Optional[str] = keytab_path
        self.redhat_api_refresh_token: Optional[str] = redhat_api_refresh_token
        self.kerberos_realm = kerberos_realm
        self.koji_build_command = koji_build_command
        if not which(pkg_tool):
            logger.warning(f"{pkg_tool} is not executable or in any path")
        self.pkg_tool: str = pkg_tool
        self.upstream_git_remote = upstream_git_remote

        self.services: set[GitService] = set()

        # %%% ACTIONS HANDLER CONFIGURATION %%%
        # these values are specific to packit service when we run actions in a sandbox
        # users will never set these, so let's hide those from them

        # name of the handler to run actions and commands, default to current env
        self.command_handler: RunCommandType = (
            RunCommandType(command_handler) if command_handler else RunCommandType.local
        )
        # a dir where the PV is mounted: both in sandbox and in worker
        self.command_handler_work_dir: str = command_handler_work_dir
        # name of the PVC so that the sandbox has the same volume mounted
        self.command_handler_pvc_env_var: str = (
            command_handler_pvc_env_var  # pointer to pointer
        )
        # name of sandbox container image
        self.command_handler_image_reference: str = command_handler_image_reference
        # do I really need to explain this?
        self.command_handler_k8s_namespace: str = command_handler_k8s_namespace
        # To be able to mount other volumes (like repository cache) to the sandcastle pod.
        # The keys are not checked by marshmallow to support any argument supported by Sandcastle.
        # e.g. you can set `path` and `pvc`/`pvc_from_env`
        self.command_handler_pvc_volume_specs: list[dict[str, str]] = (
            command_handler_pvc_volume_specs or []
        )
        # To specify PVCs' storage class (differes in auto-prod and MP+)
        self.command_handler_storage_class = command_handler_storage_class

        # Needs to be used for requesting PVCs in Sandcastle
        self.appcode = appcode

        # path to a file where OGR should store HTTP requests
        # this is used for packit testing: don't expose this to users
        self.github_requests_log_path: str = ""

        self.services = Config.load_authentication(kwargs)
        self.package_config_path = package_config_path
        self.repository_cache = repository_cache
        self.add_repositories_to_repository_cache = add_repositories_to_repository_cache
        self.default_parse_time_macros = default_parse_time_macros or {}

        # because of current load_authentication implementation it will generate false warnings
        # if kwargs:
        #     logger.warning(f"Following kwargs were not processed:" f"{kwargs}")

    def __repr__(self):
        return (
            "Config("
            f"debug='{self.debug}', "
            f"fas_user='{self.fas_user}', "
            f"keytab_path='{self.keytab_path}', "
            f"kerberos_realm='{self.kerberos_realm}', "
            f"koji_build_command='{self.koji_build_command}', "
            f"pkg_tool='{self.pkg_tool}', "
            f"upstream_git_remote='{self.upstream_git_remote}', "
            f"command_handler='{self.command_handler}', "
            f"command_handler_work_dir='{self.command_handler_work_dir}', "
            f"command_handler_pvc_env_var='{self.command_handler_pvc_env_var}', "
            f"command_handler_image_reference='{self.command_handler_image_reference}', "
            f"command_handler_k8s_namespace='{self.command_handler_k8s_namespace}', "
            f"command_handler_pvc_volume_specs='{self.command_handler_pvc_volume_specs}', "
            f"command_handler_storage_class='{self.command_handler_storage_class}', "
            f"appcode='{self.appcode}', "
            f"repository_cache='{self.repository_cache}', "
            f"default_parse_time_macros='{self.default_parse_time_macros}')"
        )

    @classmethod
    def get_user_config(cls) -> "Config":
        xdg_config_home = os.getenv("XDG_CONFIG_HOME")
        directory = (
            Path(xdg_config_home) if xdg_config_home else Path.home() / ".config"
        )

        logger.debug(f"Loading user config from directory: {directory}")

        loaded_config: dict = {}
        for config_file_name in CONFIG_FILE_NAMES:
            config_file_name_full = directory / config_file_name
            logger.debug(f"Trying to load user config from: {config_file_name_full}")
            if config_file_name_full.is_file():
                try:
                    with open(config_file_name_full) as file:
                        loaded_config = safe_load(file)
                except Exception as ex:
                    logger.error(f"Cannot load user config {config_file_name_full!r}.")
                    raise PackitException(f"Cannot load user config: {ex!r}.") from ex
                break
        return Config.get_from_dict(raw_dict=loaded_config)

    @classmethod
    def get_from_dict(cls, raw_dict: dict) -> "Config":
        # required to avoid cyclical imports
        from packit.schema import UserConfigSchema

        config = UserConfigSchema().load(raw_dict)
        logger.debug(f"Loaded config: {config}")

        return config

    @staticmethod
    def load_authentication(raw_dict):
        services = set()
        deprecated_keys = [
            "github_app_id",
            "github_app_cert_path",
            "github_token",
            "pagure_user_token",
            "pagure_instance_url",
            "pagure_fork_token",
        ]
        if "authentication" in raw_dict:
            services = get_instances_from_dict(instances=raw_dict["authentication"])
        elif any(key in raw_dict for key in deprecated_keys):
            logger.warning(
                "Please, "
                "use 'authentication' key in the user configuration "
                "to set tokens for GitHub and Pagure. "
                "New method supports more services and direct keys will be removed in the future.\n"
                "Example:\n"
                "authentication:\n"
                "    github.com:\n"
                "        token: GITHUB_TOKEN\n"
                "    pagure:\n"
                "        token: PAGURE_TOKEN\n"
                '        instance_url: "https://src.fedoraproject.org"\n'
                "See our documentation for more information "
                "http://packit.dev/docs/configuration/#user-configuration-file. ",
            )
            github_app_id = raw_dict.get("github_app_id")
            github_app_cert_path = raw_dict.get("github_app_cert_path")
            github_token = raw_dict.get("github_token")
            services.add(
                GithubService(
                    token=github_token,
                    github_app_id=github_app_id,
                    github_app_private_key_path=github_app_cert_path,
                ),
            )
            pagure_user_token = raw_dict.get("pagure_user_token")
            pagure_instance_url = raw_dict.get(
                "pagure_instance_url",
                "https://src.fedoraproject.org",
            )
            if raw_dict.get("pagure_fork_token"):
                warnings.warn(
                    "packit no longer accepts 'pagure_fork_token'"
                    " value (https://github.com/packit/packit/issues/495)",
                    stacklevel=2,
                )
            services.add(
                PagureService(
                    token=pagure_user_token,
                    instance_url=pagure_instance_url,
                ),
            )

        return services

    def _get_project(
        self,
        url: str,
        required: bool = True,
        get_project_kwargs: Optional[dict] = None,
    ) -> Optional[GitProject]:
        """
        Gets a GitProject for the given URL.

        Args:
            url: Project URL.
            required: Whether to raise an exception on failure or return None.
            get_project_kwargs: Keyword arguments to be passed to ogr's get_project().

        Returns:
            GitProject instance or None if the underlying get_project() call fails
            and required is False.

        Raises:
            PackitConfigException if the underlying get_project() call fails
            and required is True.
        """
        get_project_kwargs = get_project_kwargs or {}
        try:
            project = get_project(
                url=url,
                custom_instances=self.services,
                **get_project_kwargs,
            )
        except OgrException as ex:
            msg = f"Authentication for url {url!r} is missing in the config."
            logger.warning(msg)
            if required:
                raise PackitConfigException(msg) from ex
            return None
        return project

    def get_project(
        self,
        url: str,
        required: bool = True,
        get_project_kwargs: Optional[dict] = None,
    ) -> Proxy:
        """
        Gets a proxy of GitProject for the given URL. On access, if the underlying
        get_project() call fails, the behaviour depends on the `required` argument.
        If set to True (the default), a PackitConfigException is raised, otherwise
        the proxy acts as it was None.

        Args:
            url: Project URL.
            required: Whether to raise an exception on failure or act as None.
            get_project_kwargs: Keyword arguments to be passed to ogr's get_project().

        Returns:
            Proxy of a GitProject instance or None.
        """
        return Proxy(partial(self._get_project, url, required, get_project_kwargs))


pass_config = click.make_pass_decorator(Config)


def get_default_map_from_file() -> Optional[dict]:
    config_path = Path(".packit")
    if config_path.is_file():
        return json.loads(config_path.read_text())
    return None


@lru_cache
def get_context_settings() -> dict:
    return {
        "help_option_names": ["-h", "--help"],
        "auto_envvar_prefix": "PACKIT",
        "default_map": get_default_map_from_file(),
    }


class RunCommandType(Enum):
    sandcastle = "sandcastle"
    local = "local"



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest

from packit.utils.source_script import create_source_script


@pytest.mark.parametrize(
    "ref, pr_id, merge_pr, target_branch, job_config_index, url, command",
    [
        (
            None,
            None,
            True,
            None,
            None,
            "https://github.com/packit/ogr",
            'packit -d prepare-sources --result-dir "$resultdir" --no-create-symlinks '
            "https://github.com/packit/ogr",
        ),
        (
            "123",
            None,
            True,
            None,
            None,
            "https://github.com/packit/ogr",
            'packit -d prepare-sources --result-dir "$resultdir" --ref 123 --no-create-symlinks '
            "https://github.com/packit/ogr",
        ),
        (
            None,
            "1",
            False,
            None,
            None,
            "https://github.com/packit/ogr",
            'packit -d prepare-sources --result-dir "$resultdir" --pr-id 1 '
            "--no-merge-pr --no-create-symlinks https://github.com/packit/ogr",
        ),
        (
            None,
            "1",
            True,
            "main",
            None,
            "https://github.com/packit/ogr",
            'packit -d prepare-sources --result-dir "$resultdir" --pr-id 1 '
            "--merge-pr --target-branch main --no-create-symlinks https://github.com/packit/ogr",
        ),
        (
            None,
            "1",
            True,
            "main",
            0,
            "https://github.com/packit/ogr",
            'packit -d prepare-sources --result-dir "$resultdir" --pr-id 1 '
            "--merge-pr --target-branch main --job-config-index 0 "
            "--no-create-symlinks https://github.com/packit/ogr",
        ),
    ],
)
def test_create(ref, pr_id, merge_pr, target_branch, job_config_index, url, command):
    assert (
        create_source_script(
            ref=ref,
            pr_id=pr_id,
            merge_pr=merge_pr,
            target_branch=target_branch,
            job_config_index=job_config_index,
            url=url,
        )
        == f"""
#!/bin/sh

git config --global user.email "hello@packit.dev"
git config --global user.name "Packit"
resultdir=$PWD
{command}

"""
    )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest
from flexmock import flexmock
from gnupg import GPG

from packit.api import PackitAPI
from packit.exceptions import PackitException
from packit.security import CommitVerifier
from tests.integration.utils import remove_gpg_key_pair


def test_allowed_gpg_keys_none(api_instance_source_git: PackitAPI):
    api_instance_source_git.up.allowed_gpg_keys = None
    flexmock(CommitVerifier).should_receive("check_signature_of_commit").times(0)
    api_instance_source_git.up.check_last_commit()


@pytest.mark.parametrize("allowed_gpg_keys", [[], ["abcd", "efgh"]])
def test_allowed_gpg_keys_not_allowed(
    allowed_gpg_keys,
    api_instance_source_git: PackitAPI,
    gnupg_key_fingerprint: str,
):
    api_instance_source_git.up.local_project.git_repo.git.commit(
        message="signed commit",
        gpg_sign=gnupg_key_fingerprint,
        allow_empty=True,
    )

    api_instance_source_git.up.allowed_gpg_keys = allowed_gpg_keys
    with pytest.raises(PackitException) as ex:
        api_instance_source_git.up.check_last_commit()
    assert "not signed" in str(ex)


def test_allowed_gpg_keys_allowed(
    api_instance_source_git: PackitAPI,
    gnupg_key_fingerprint: str,
):
    api_instance_source_git.up.local_project.git_repo.git.commit(
        message="signed commit",
        gpg_sign=gnupg_key_fingerprint,
        allow_empty=True,
    )

    api_instance_source_git.up.allowed_gpg_keys = [gnupg_key_fingerprint]
    api_instance_source_git.up.check_last_commit()


def test_allowed_gpg_keys_not_existing_key(
    api_instance_source_git: PackitAPI,
    gnupg_instance: GPG,
    gnupg_key_fingerprint: str,
):
    api_instance_source_git.up.local_project.git_repo.git.commit(
        message="signed commit",
        gpg_sign=gnupg_key_fingerprint,
        allow_empty=True,
    )
    remove_gpg_key_pair(
        gpg_binary=gnupg_instance.gpgbinary,
        fingerprint=gnupg_key_fingerprint,
    )
    api_instance_source_git.up.allowed_gpg_keys = [gnupg_key_fingerprint]
    with pytest.raises(PackitException) as ex:
        api_instance_source_git.up.check_last_commit()
    assert "Cannot receive a gpg key" in str(ex)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
Update a dist-git repo from a source-git repo
"""

import pathlib
from shutil import which
from typing import Optional

import click

from packit.api import PackitAPI
from packit.cli.utils import cover_packit_exception, iterate_packages_source_git
from packit.config import Config, PackageConfig, pass_config
from packit.local_project import CALCULATE, LocalProjectBuilder


@click.command("update-dist-git")
@click.option(
    "--upstream-ref",
    type=str,
    help="Git ref of the last upstream commit in the current branch "
    "from which packit should generate patches "
    "(this option implies the repository is source-git).",
)
@click.option(
    "--pkg-tool",
    type=str,
    help="""Name or path of the packaging tool used to work with
    sources in the dist-git repo. A variant of 'rpkg'.

    Skip retrieving and uploading source archives to the lookaside cache
    if not specified.
    """,
)
@click.option(
    "-m",
    "--message",
    metavar="<msg>",
    type=str,
    help="""Commit the changes in the dist-git repository and use <msg>
    as the commit message.

    Mutually exclusive with -F.""",
)
@click.option(
    "-F",
    "--file",
    metavar="<file>",
    type=click.Path(exists=True, dir_okay=False, allow_dash=True),
    help="""Commit the changes in the dist-git repository and take the commit message
    from <file>. Use - to read from the standard input.""",
)
@click.option(
    "-f",
    "--force",
    default=False,
    is_flag=True,
    help="Don't check the synchronization status of the source-git"
    " and dist-git repos prior to performing the update.",
)
@click.argument("source-git", type=click.Path(exists=True, file_okay=False))
@click.argument("dist-git", type=click.Path(exists=True, file_okay=False))
@pass_config
@cover_packit_exception
@iterate_packages_source_git
def update_dist_git(
    config: Config,
    package_config: PackageConfig,
    source_git: str,
    dist_git: str,
    upstream_ref: Optional[str],
    pkg_tool: Optional[str],
    message: Optional[str],
    file: Optional[str],
    force: Optional[bool],
):
    """Update a dist-git repository using content from a source-git repository

    Update a dist-git repository with patches created from the commits
    between <upstream_ref> and the current HEAD of the source-git repo.

    This command, by default, performs only local operations and uses
    the content of the source-git and dist-git repository as it is:
    does not checkout branches or fetches remotes.

    A commit in dist-git is created only if a commit message is provided with
    --message or --file. This commit will have a 'From-source-git-commit'
    Git-trailer appended to it, to mark the hash of the source-git commit
    from which it is created.

    The source archives are retrieved from the upstream URLs specified in
    the spec-file and uploaded to the lookaside cache in dist-git only if
    '--pkg-tool' is specified.

    Examples:

    To update a dist-git repo from source-git without uploading the source-archive
    to the lookaside cache and creating a commit with the updates, run:

    \b
        $ packit source-git update-dist-git src/curl rpms/curl


    To also commit the changes and upload the source-archive to the lookaside-cache
    specify -m and --pkg-tool:

    \b
        $ packit source-git update-dist-git -m'Update from source-git' \\
                --pkg-tool fedpkg src/curl rpms/curl
    """
    if message and file:
        raise click.BadOptionUsage("-m", "Option -m cannot be combined with -F.")
    if pkg_tool and not which(pkg_tool):
        raise click.BadOptionUsage(
            "--pkg-tool",
            f"{pkg_tool} is not executable or in any path",
        )
    if file:
        with click.open_file(file, "r") as fp:
            message = fp.read()

    source_git_path = pathlib.Path(source_git).resolve()
    dist_git_path = pathlib.Path(dist_git).resolve()
    builder = LocalProjectBuilder(offline=True)
    api = PackitAPI(
        config=config,
        package_config=package_config,
        upstream_local_project=builder.build(
            working_dir=source_git_path,
            git_repo=CALCULATE,
        ),
        downstream_local_project=builder.build(
            working_dir=dist_git_path,
            git_repo=CALCULATE,
        ),
    )

    title, _, message = message.partition("\n\n") if message else (None, None, None)
    api.update_dist_git(
        version=None,
        upstream_ref=upstream_ref or package_config.upstream_ref,
        # Add new sources if a pkg_tool was specified.
        add_new_sources=bool(pkg_tool),
        force_new_sources=False,
        upstream_tag=None,
        commit_title=title,
        commit_msg=message,
        sync_default_files=False,
        pkg_tool=pkg_tool,
        mark_commit_origin=True,
        check_sync_status=not force,
    )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
Functional tests for local-build command
"""
import re
from pathlib import Path
from subprocess import CalledProcessError

import pytest
from distro import linux_distribution

from tests.functional.spellbook import call_real_packit


def test_rpm_command(ogr_distgit_and_remote):
    call_real_packit(
        parameters=["--debug", "build", "locally"],
        cwd=ogr_distgit_and_remote[0],
    )
    rpm_paths = ogr_distgit_and_remote[0].glob("noarch/*.rpm")

    assert all(rpm_path.exists() for rpm_path in rpm_paths)


def test_local_build_with_remote_good(ogr_distgit_and_remote):
    call_real_packit(
        parameters=["--debug", "--remote", "origin", "build", "locally"],
        cwd=ogr_distgit_and_remote[0],
    )
    rpm_paths = ogr_distgit_and_remote[0].glob("noarch/*.rpm")

    assert all(rpm_path.exists() for rpm_path in rpm_paths)


def test_local_build_with_remote_bad(ogr_distgit_and_remote):
    with pytest.raises(CalledProcessError) as ex:
        call_real_packit(
            parameters=["--debug", "--remote", "burcak", "build", "locally"],
            cwd=ogr_distgit_and_remote[0],
            return_output=True,
        )
    assert b"Remote named 'burcak' didn't exist" in ex.value.output


@pytest.mark.skipif(
    linux_distribution()[0].startswith("CentOS"),
    reason="No rpmautospec-rpm-macros installed",
)
def test_rpm_command_for_path(ogr_distgit_and_remote):
    call_real_packit(
        parameters=["--debug", "build", "locally", str(ogr_distgit_and_remote[0])],
    )
    rpm_paths = Path.cwd().glob("noarch/*.rpm")
    assert all(rpm_path.exists() for rpm_path in rpm_paths)


def test_local_build_from_srpm(ogr_distgit_and_remote):
    call_real_packit(parameters=["--debug", "srpm"], cwd=ogr_distgit_and_remote[0])

    the_chosen_srpm = next(ogr_distgit_and_remote[0].glob("*.src.rpm"))

    call_real_packit(
        parameters=["--debug", "build", "--srpm", the_chosen_srpm, "locally"],
        cwd=ogr_distgit_and_remote[0],
    )
    rpm_paths = ogr_distgit_and_remote[0].glob("noarch/*.rpm")

    # we need to be able to find a part of version and release in the built RPMs
    # if not ⇒ we ran RPM build from the CWD
    version_release = re.search(r"ogr-(.*)\.main.*", str(the_chosen_srpm)).group(1)
    assert all(version_release in str(path) for path in rpm_paths)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
Functions and classes dealing with syncing files between repositories
"""
import copy
import glob
import logging
import os
from collections.abc import Iterator, Sequence
from pathlib import Path
from typing import Optional, Union

from packit.exceptions import PackitException
from packit.utils import run_command

logger = logging.getLogger(__name__)


def check_subpath(
    subpath: Path,
    path: Path,
    ensure_trailing_slash: bool = False,
) -> str:
    """Check if 'subpath' is a subpath of 'path'

    Args:
        subpath: Subpath to be checked.
        path: Path agains which subpath is checked.

    Returns:
        'subpath', resolved, in case it is a subpath of 'path'.

    Raises:
        PackitException, if 'subpath' is not a subpath of 'path'.
    """
    try:
        subpath.resolve().relative_to(path.resolve())
    except ValueError as e:
        raise PackitException(
            f"Sync files: Illegal path! {subpath} is not in the subpath of {path}.",
        ) from e
    ret = str(subpath.resolve())
    if ensure_trailing_slash:
        ret += os.sep
    return ret


class SyncFilesItem:
    """Some files to sync to destination

    Think about this as a wrapper around 'rsync'.

    Attributes:
        src: List of paths to sync.
        dest: Destination to sync to.
        mkpath: Create the destination's path component.
        delete: Delete extra files from dest dirs.
        filters: List of rsync filters used for syncing.
    """

    def __init__(
        self,
        src: Sequence[Union[str, Path]],
        dest: Union[str, Path],
        mkpath: bool = False,
        delete: bool = False,
        filters: Optional[list[str]] = None,
    ):
        # pathlib.Path has no support for trailing slashes, but
        # a trailing slash has a meaning for rsync.
        # Store all paths as strings internally to keep trailing
        # slashes.
        self.src = [str(s) for s in src]
        self.dest = str(dest)
        self.mkpath = mkpath
        self.delete = delete
        self.filters = filters or []

    def __repr__(self):
        return (
            f"SyncFilesItem(src={self.src}, "
            f"dest={self.dest}, "
            f"mkpath={self.mkpath}, "
            f"delete={self.delete}, "
            f"filters={self.filters})"
        )

    def __str__(self):
        return " ".join(self.command())

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, SyncFilesItem):
            raise NotImplementedError()

        return (self.src, self.dest, self.mkpath, self.delete, self.filters) < (
            other.src,
            other.dest,
            other.mkpath,
            other.delete,
            other.filters,
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SyncFilesItem):
            raise NotImplementedError()

        return (self.src, self.dest, self.mkpath, self.delete, self.filters) == (
            other.src,
            other.dest,
            other.mkpath,
            other.delete,
            other.filters,
        )

    def command(self, fail_on_missing: bool = False) -> list[str]:
        """Provide to command to do the sync

        Args:
            fail_on_missing: Flag to make the command fail if any of
                the sources are missing.

        Returns:
            The command to do the sync, as a list of strings.
        """
        command = ["rsync", "--archive"]
        if self.delete:
            command += ["--delete"]
        for filter_ in self.filters:
            command += ["--filter", filter_]
        if self.mkpath:
            command += ["--mkpath"]
        if not fail_on_missing:
            command += ["--ignore-missing-args"]
        for src in self.src:
            globs = glob.glob(src)
            if globs:
                command += globs
            else:
                # 'globs' is an empty list if 'src' is missing, which could
                # render the 'rsync' command meaningless.
                # Make sure 'src' is part of the command in these cases,
                # and let --ignore-missing-args handle the rest.
                command += [src]
        command += [self.dest]
        return command

    def resolve(
        self,
        src_base: Optional[Path] = None,
        dest_base: Optional[Path] = None,
    ):
        """Resolve all paths and check they are relative to src_base and dest_base

        Args:
            src_base: Base directory for all src items.
            dest_base: Base directory for dest.
        """
        src_base = src_base or Path.cwd()
        dest_base = dest_base or Path.cwd()
        self.src = [
            check_subpath(src_base / path, src_base, path.endswith(os.sep))
            for path in self.src
        ]
        self.dest = check_subpath(
            dest_base / self.dest,
            dest_base,
            self.dest.endswith(os.sep),
        )

    def drop_src(
        self,
        src: Union[str, Path],
        criteria=lambda x, y: x == str(y),
    ) -> Optional["SyncFilesItem"]:
        """Remove 'src' from the list of src-s

        Args:
            src: A path to be removed.
            criteria: Function to tell if a src should be removed.
                Receives two arguments: the src item inspected and
                'src' received by this method.
        Returns:
            A *new* SyncFilesItem instance if the internal src list
            has still some items left after 'src' is removed.
            Otherwise returns None.
        """
        new_src = [s for s in self.src if not criteria(s, src)]
        if new_src:
            self_copy = copy.copy(self)
            self_copy.src = new_src
            return self_copy
        return None


def iter_srcs(synced_files: Sequence[SyncFilesItem]) -> Iterator[str]:
    """Iterate over all the src-s in a list of SyncFilesItem

    Args:
        synced_files: List of SyncFilesItem.

    Yields:
        src-s from every SyncFilesItem, one by one.
    """
    for item in synced_files:
        yield from item.src


def sync_files(synced_files: Sequence[SyncFilesItem]):
    """
    Copy files b/w upstream and downstream repo.
    """
    for item in synced_files:
        command = item.command()
        logger.debug(f"Running {command!r} ...")
        run_command(command, print_live=True)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
E2E tests which utilize cockpit projects
"""

import pytest

from packit.cli.utils import get_packit_api
from packit.local_project import CALCULATE, LocalProjectBuilder
from packit.utils.commands import cwd
from tests.spellbook import (
    DG_OGR,
    UP_EDD,
    UP_VSFTPD,
    build_srpm,
    get_test_config,
    initiate_git_repo,
)


@pytest.fixture(
    params=[
        (UP_EDD, "0.3", "https://github.com/psss/edd"),
        (UP_VSFTPD, "3.0.3", "https://github.com/olysonek/vsftpd"),
        (DG_OGR, None, "https://src.fedoraproject.org/rpms/python-ogr"),
    ],
    ids=["edd", "vsftpd", "ogr"],
)
def example_repo(request, tmp_path):
    example_path, tag, remote = request.param
    u = tmp_path / "up"
    initiate_git_repo(u, tag=tag, copy_from=example_path, upstream_remote=remote)
    return u


def test_srpm_on_example(example_repo):
    c = get_test_config()
    api = get_packit_api(
        config=c,
        local_project=LocalProjectBuilder().build(
            working_dir=example_repo,
            git_repo=CALCULATE,
        ),
    )
    with cwd(example_repo):
        path = api.create_srpm()
    assert path.exists()
    build_srpm(path)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import os
import subprocess

import pytest
from ogr.services.pagure import PagureService

from tests.spellbook import git_set_user_email


@pytest.mark.skipif(
    condition=True,
    reason="Don't interact with a real pagure instance by default",
)
def test_basic_distgit_workflow(tmp_path):
    pagure_token = os.getenv("PAGURE_TOKEN")

    pag = PagureService(
        token=pagure_token,
        repo="tmux-top",
        namespace="rpms",
        instance_url="https://src.stg.fedoraproject.org/",
    )

    print(pag.pagure.whoami())

    proj = pag.get_project()

    fork = proj.get_fork(create=True)
    clone_url = fork.get_git_urls()["ssh"]

    repo = tmp_path.joinpath("repo")

    branch_name = "cookie"

    subprocess.check_call(["git", "clone", clone_url, repo])
    git_set_user_email(repo)
    subprocess.check_call(["git", "checkout", "-B", branch_name], cwd=repo)

    repo.joinpath("README").write_text("just trying something out\n")

    subprocess.check_call(["git", "add", "README"], cwd=repo)
    subprocess.check_call(["git", "commit", "-m", "test commit"], cwd=repo)
    subprocess.check_call(
        ["git", "push", "origin", f"{branch_name}:{branch_name}"],
        cwd=repo,
    )

    pr = fork.create_pr("testing PR", "serious description", "master", branch_name)

    pr.comment("howdy!")



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest

from packit.cli.prepare_sources import load_job_config
from packit.config import CommonPackageConfig, JobConfig, JobConfigTriggerType, JobType


@pytest.mark.parametrize(
    "raw_job_config,expected",
    [
        pytest.param(
            """\
            {
                "job": "copr_build",
                "trigger": "release",
                "skip_build": false,
                "manual_trigger": false,
                "packages": {
                    "package": {
                        "upstream_package_name": null,
                        "downstream_package_name": "package",
                        "archive_root_dir_template": "{upstream_pkg_name}-{version}",
                        "patch_generation_ignore_paths": [],
                        "dist_git_namespace": "rpms",
                        "actions": {},
                        "upstream_project_url": null,
                        "sources": [],
                        "create_pr": true,
                        "merge_pr_in_ci": true,
                        "upstream_ref": null,
                        "upstream_tag_template": "{version}",
                        "config_file_path": null,
                        "synced_files": [],
                        "sync_changelog": false,
                        "patch_generation_patch_id_digits": 4,
                        "preserve_project": false,
                        "branch": null,
                        "additional_repos": [],
                        "env": {},
                        "additional_packages": [],
                        "scratch": false,
                        "targets": {},
                        "fmf_url": null,
                        "fmf_ref": null,
                        "fmf_path": null,
                        "owner": null,
                        "use_internal_tf": false,
                        "list_on_homepage": false,
                        "project": "example1",
                        "dist_git_branches": [],
                        "timeout": 7200,
                        "notifications": {
                            "pull_request": {
                                "successful_build": false
                            }
                        },
                        "copy_upstream_release_description": false,
                        "specfile_path": "package.spec",
                        "dist_git_base_url": "https://src.fedoraproject.org/",
                        "spec_source_id": "Source0",
                        "allowed_gpg_keys": null,
                        "tmt_plan": null,
                        "tf_post_install_script": null
                    }
                }
            }
            """,
            JobConfig(
                type=JobType.copr_build,
                trigger=JobConfigTriggerType.release,
                packages={
                    "package": CommonPackageConfig(
                        downstream_package_name="package",
                        specfile_path="package.spec",
                        project="example1",
                    ),
                },
            ),
            id="valid",
        ),
        pytest.param('{"config_file_path": null}', None, id="invalid"),
    ],
)
def test_load_job_config(raw_job_config, expected):
    assert load_job_config(raw_job_config) == expected



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import shutil
import subprocess
from pathlib import Path

import pytest
from deepdiff import DeepDiff
from flexmock import flexmock

from packit.config import JobConfig, PackageConfig
from packit.utils.commands import cwd
from packit.utils.repo import create_new_repo
from tests.spellbook import (
    DG_OGR,
    DISTGIT,
    DISTGIT_WITH_AUTOCHANGELOG,
    UPSTREAM,
    UPSTREAM_SPEC_NOT_IN_ROOT,
    UPSTREAM_WEIRD_SOURCES,
    UPSTREAM_WITH_MUTLIPLE_SOURCES,
    initiate_git_repo,
    prepare_dist_git_repo,
)


def get_git_repo_and_remote(
    target_dir: Path,
    repo_template_path: Path,
) -> tuple[Path, Path]:
    """
    :param target_dir: tmpdir from pytest - we'll work here
    :param repo_template_path: git repo template from tests/data/
    """
    u_remote_path = target_dir / f"upstream_remote-{repo_template_path.name}"
    u_remote_path.mkdir(parents=True, exist_ok=True)
    create_new_repo(u_remote_path, ["--bare"])

    u = target_dir / f"local_clone-{repo_template_path.name}"
    shutil.copytree(repo_template_path, u)
    initiate_git_repo(u, tag="0.1.0", push=True, upstream_remote=str(u_remote_path))

    return u, u_remote_path


@pytest.fixture()
def upstream_and_remote(tmp_path) -> tuple[Path, Path]:
    return get_git_repo_and_remote(tmp_path, UPSTREAM)


@pytest.fixture()
def upstream_and_remote_with_multiple_sources(tmp_path) -> tuple[Path, Path]:
    return get_git_repo_and_remote(tmp_path, UPSTREAM_WITH_MUTLIPLE_SOURCES)


@pytest.fixture()
def upstream_and_remote_weird_sources(tmp_path) -> tuple[Path, Path]:
    return get_git_repo_and_remote(tmp_path, UPSTREAM_WEIRD_SOURCES)


@pytest.fixture()
def upstream_spec_not_in_root(tmp_path) -> tuple[Path, Path]:
    return get_git_repo_and_remote(tmp_path, UPSTREAM_SPEC_NOT_IN_ROOT)


@pytest.fixture()
def distgit_and_remote(tmp_path) -> tuple[Path, Path]:
    d_remote_path = tmp_path / "dist_git_remote"
    d_remote_path.mkdir(parents=True, exist_ok=True)
    create_new_repo(d_remote_path, ["--bare"])

    d = tmp_path / "dist_git"
    shutil.copytree(DISTGIT, d)
    initiate_git_repo(
        d,
        push=True,
        remotes=[
            ("origin", str(d_remote_path)),
            ("i_am_distgit", "https://src.fedoraproject.org/rpms/python-ogr"),
        ],
    )
    prepare_dist_git_repo(d)

    return d, d_remote_path


@pytest.fixture()
def distgit_with_autochangelog_and_remote(tmp_path) -> tuple[Path, Path]:
    d_remote_path = tmp_path / "autochangelog_dist_git_remote"
    d_remote_path.mkdir(parents=True, exist_ok=True)
    create_new_repo(d_remote_path, ["--bare"])

    d = tmp_path / "autochangelog_dist_git"
    shutil.copytree(DISTGIT_WITH_AUTOCHANGELOG, d)
    initiate_git_repo(
        d,
        push=True,
        remotes=[
            ("origin", str(d_remote_path)),
            ("i_am_distgit", "https://src.fedoraproject.org/rpms/python-ogr"),
        ],
    )
    prepare_dist_git_repo(d)

    return d, d_remote_path


@pytest.fixture()
def ogr_distgit_and_remote(tmp_path) -> tuple[Path, Path]:
    d_remote_path = tmp_path / "ogr_dist_git_remote"
    d_remote_path.mkdir(parents=True, exist_ok=True)
    create_new_repo(d_remote_path, ["--bare"])

    d = tmp_path / "ogr_dist_git"
    shutil.copytree(DG_OGR, d)
    initiate_git_repo(
        d,
        push=True,
        remotes=[
            ("origin", str(d_remote_path)),
            ("i_am_distgit", "https://src.fedoraproject.org/rpms/python-ogr"),
        ],
    )
    prepare_dist_git_repo(d)
    return d, d_remote_path


@pytest.fixture(params=["upstream", "ogr-distgit"])
def upstream_or_distgit_path(
    request,
    upstream_and_remote,
    distgit_and_remote,
    ogr_distgit_and_remote,
):
    """
    Parametrize the test to upstream, downstream [currently skipped] and ogr distgit
    """
    return {
        "upstream": upstream_and_remote[0],
        "distgit": distgit_and_remote[0],
        "ogr-distgit": ogr_distgit_and_remote[0],
    }[request.param]


@pytest.fixture(
    params=["upstream", "distgit", "ogr-distgit", "upstream-with-multiple-sources"],
)
def cwd_upstream_or_distgit(
    request,
    upstream_and_remote,
    distgit_and_remote,
    ogr_distgit_and_remote,
    upstream_and_remote_with_multiple_sources,
):
    """
    Run the code from upstream, downstream and ogr-distgit.

    When using be careful to
        - specify this fixture in the right place
        (the order of the parameters means order of the execution)
        - to not overwrite the cwd in the other fixture or in the test itself
    """
    cwd_path = {
        "upstream": upstream_and_remote[0],
        "distgit": distgit_and_remote[0],
        "ogr-distgit": ogr_distgit_and_remote[0],
        "upstream-with-multiple-sources": upstream_and_remote_with_multiple_sources[0],
    }[request.param]

    with cwd(cwd_path):
        yield cwd_path


@pytest.fixture
def copr_client_mock(get_list_return=None):
    get_list_return = {
        "centos-stream-aarch64": "",
        "custom-1-x86_64": "",
        "epel-6-x86_64": "",
        "epel-8-x86_64": "",
        "fedora-31-aarch64": "",
        "fedora-31-armhfp": "This is emulated on x86_64",
        "fedora-31-x86_64": "",
        "fedora-32-armhfp": "This is emulated on x86_64",
        "fedora-32-i386": "Not-released Koji packages",
        "fedora-32-s390x": "This is emulated on x86_64",
        "fedora-32-x86_64": "",
        "fedora-33-x86_64": "",
        "fedora-eln-x86_64": "",
        "fedora-rawhide-aarch64": "",
        "fedora-rawhide-x86_64": "",
    }

    return flexmock(mock_chroot_proxy=flexmock(get_list=lambda: get_list_return))


@pytest.fixture(autouse=True, scope="function")
def configure_git():
    CMDS = [
        ["git", "config", "--global", "user.email", "packit@redhat.com"],
        ["git", "config", "--global", "user.name", "Packit Test"],
    ]
    # verify that git is already configured
    try:
        output = subprocess.check_output(["git", "config", "-l", "--global"])
    except subprocess.CalledProcessError:
        output = ""
    if "user.name" in output if isinstance(output, str) else output.decode():
        return
    for item in CMDS:
        subprocess.call(item)


def pytest_assertrepr_compare(op, left, right):
    """Use DeepDiff to show the difference between dictionaries and config objects
    when they are not equal.

    Makes figuring out test failures somewhat easier.
    """
    if isinstance(left, JobConfig) and isinstance(right, JobConfig) and op == "==":
        from packit.schema import JobConfigSchema

        schema = JobConfigSchema()
        return [str(DeepDiff(schema.dump(left), schema.dump(right)))]
    if (
        isinstance(left, PackageConfig)
        and isinstance(right, PackageConfig)
        and op == "=="
    ):
        from packit.schema import PackageConfigSchema

        schema = PackageConfigSchema()
        return [str(DeepDiff(schema.dump(left), schema.dump(right)))]
    if isinstance(left, dict) and isinstance(right, dict) and op == "==":
        return [str(DeepDiff(left, right))]
    return None



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging
import os
import shlex
import subprocess
import sys
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Union

from packit.exceptions import PackitCommandFailedError
from packit.utils.logging import StreamLogger

logger = logging.getLogger(__name__)


@dataclass
class CommandResult:
    """
    Structure to represent a result of the command that was run.

    Attributes:
        success: Boolean value holding a result of the command.
        stdout: Holds standard output of the command, in case it was requsted.
        stderr: Holds standard error output of the command, in case it was requsted.
    """

    success: bool = False
    stdout: Optional[str] = None
    stderr: Optional[str] = None


def run_command(
    cmd: Union[list[str], str],
    error_message: Optional[str] = None,
    cwd: Union[str, Path, None] = None,
    fail: bool = True,
    output: bool = False,
    env: Optional[dict] = None,
    print_live: bool = False,
) -> CommandResult:
    """
    Run provided command in a new subprocess.

    Args:
        cmd: Command to be run.
        error_message: Error message to be included in the exception and logs in
            case the command fails.

            Defaults to generic message with the escaped command.
        cwd: Working directory of the new subprocess.

            Defaults to current working directory of the process itself.
        fail: Raise an exception if the command fails.

            Defaults to `True`.
        output: Return the output of the subprocess.

            Defaults to `False`.
        env: Environment variables to be set in the newly created subprocess.

            Defaults to none.
        print_live: Print real-time output of the command as INFO.

            Defaults to `False`.

    Returns:
        Data structure holding a result of the run command. In case it was
        requested to keep the output, they are provided as decoded strings in the
        data structure.
    """
    if not isinstance(cmd, list):
        cmd = shlex.split(cmd)

    escaped_command = " ".join(cmd)

    logger.debug(f"Command: {escaped_command}")

    cwd = str(cwd) if cwd else str(Path.cwd())
    error_message = error_message or f"Command {escaped_command!r} failed."

    # we need to pass complete env to Popen, otherwise we lose everything from os.environ
    cmd_env = os.environ
    if env:
        cmd_env.update(env)

    # we can't use universal newlines here b/c the output from the command can be encoded
    # in something alien and we would "can't decode this using utf-8" errors
    # https://github.com/packit/systemd-rhel8-flock/pull/9#issuecomment-550184016
    shell = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=False,
        cwd=cwd,
        env=cmd_env,
    )

    stdout = StreamLogger(
        shell.stdout,
        log_level=logging.DEBUG if not print_live else logging.INFO,
        decode=True,
    )
    stderr = StreamLogger(
        shell.stderr,
        log_level=logging.DEBUG if not print_live else logging.INFO,
        decode=True,
    )

    stdout.start()
    stderr.start()
    shell.wait()
    stdout.join()
    stderr.join()

    success = True  # default is success
    if shell.returncode != 0:
        logger.error(f"{error_message}")
        if fail:
            stderr_output = stderr.get_output().decode()
            stdout_output = stdout.get_output().decode()
            if output:
                logger.debug(f"Command stderr: {stderr_output}")
                logger.debug(f"Command stdout: {stdout_output}")
            raise PackitCommandFailedError(
                f"{error_message}",
                stdout_output=stdout_output,
                stderr_output=stderr_output,
            )
        success = False

    if not output:
        return CommandResult(success=success)

    command_output = (
        out.get_output().decode(sys.getdefaultencoding()) for out in (stdout, stderr)
    )

    out, err = command_output
    return CommandResult(success, out, err)


def run_command_remote(
    cmd,
    error_message=None,
    cwd=None,
    fail=True,
    output=False,
    env: Optional[dict] = None,
    print_live: bool = False,
):
    """
    Wrapper for the `run_command` method.

    Indicating that this command run some action without local effect,
    or the effect is not important.

    For example submit something to some server, and check how server reply, e.g.
    call kinit to obtain a ticket.
    """
    return run_command(
        cmd,
        error_message,
        cwd,
        fail,
        output,
        env,
        print_live=print_live,
    )


@contextmanager
def cwd(target: Union[str, Path]):
    """
    Manage cwd in a pushd/popd fashion.

    Usage:

        with cwd(tmpdir):
          do something in tmpdir
    """
    curdir = os.getcwd()
    os.chdir(target)
    try:
        yield
    finally:
        os.chdir(curdir)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT
import pytest
import rpm
from specfile import Specfile

spec = """
Name: bring-me-to-the-life
Version: 1.0
Release: 1
Source0: foo.bar
License: GPLv3+
Summary: evanescence
%description
-
%changelog
"""


@pytest.mark.parametrize(
    "spec_content, has_autochangelog",
    [
        pytest.param(
            spec
            + """
%autochangelog
""",
            True,
        ),
        pytest.param(
            spec
            + """
* Mon Mar 04 2019 Foo Bor <foo-bor@example.com> - 1.0-1
- Initial package.
""",
            False,
        ),
        pytest.param(
            spec
            + """
 # %autochangelog
""",
            False,
        ),
    ],
)
@pytest.mark.skipif(
    rpm.__version__ < "4.16",
    reason="%autochangelog requires rpm 4.16 or higher",
)
def test_set_spec_has_autochangelog(spec_content, has_autochangelog, tmp_path):
    spec_path = tmp_path / "life.spec"
    spec_path.write_text(spec_content)
    specfile = Specfile(spec_path, sourcedir=tmp_path, autosave=True)

    assert specfile.has_autochangelog == has_autochangelog



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT


"""
Update a source-git repo from a dist-git repo
"""

import pathlib

import click

from packit.api import PackitAPI
from packit.cli.utils import cover_packit_exception, iterate_packages_source_git
from packit.config import Config, PackageConfig, pass_config
from packit.local_project import CALCULATE, LocalProjectBuilder


@click.command("update-source-git")
@click.option(
    "-f",
    "--force",
    default=False,
    is_flag=True,
    help="Don't check the synchronization status of the source-git"
    " and dist-git repos prior to performing the update.",
)
@click.argument("dist-git", type=click.Path(exists=True, file_okay=False))
@click.argument("source-git", type=click.Path(exists=True, file_okay=False))
@click.argument("revision-range", required=False)
@pass_config
@cover_packit_exception
@iterate_packages_source_git
def update_source_git(
    config: Config,
    package_config: PackageConfig,
    source_git: str,
    dist_git: str,
    revision_range: str,
    force: bool,
):
    """Update a source-git repository based on a dist-git repository.

    Update a source-git repository with the selected checkout of a spec file
    and additional packaging files from a dist-git repository.

    Revision range represents part of dist-git history which is supposed
    to be synchronized. Use `HEAD~..` if you want to synchronize the last
    commit from dist-git. For more information on possible revision range
    formats, see gitrevisions(7). If the revision range is not specified,
    dist-git commits with no counterpart in source-git will be synchronized.

    If patches or the sources file in the spec file changed, the command
    exits with return code 2. Such changes are not supported by this
    command, code changes should happen in the source-git repo.

    Inapplicable changes to the .gitignore file are ignored since the
    file may not be synchronized between dist-git and source-git.

    This command, by default, performs only local operations and uses the
    content of the source-git and dist-git repositories as it is, no checkout
    or fetch is performed.

    After the synchronization is done, packit will inform about the changes
    it has performed and about differences between source-git and dist-git
    prior to the synchronization process.

    Dist-git commit messages are preserved and used when creating new
    source-git commits, but a 'From-dist-git-commit' trailer is appended
    to them to mark the hash of the dist-git commit from which they
    are created.


    Examples

    Take the extra (not synchronized) commit(s) of systemd dist-git repo and
    copy the spec file and other packaging files into the source-git repo:

    \b
        $ packit source-git update-source-git rpms/systemd src/systemd

    Synchronize changes from the last three dist-git commits:

    \b
        $ packit source-git update-source-git rpms/systemd src/systemd HEAD~3..
    """
    if force and not revision_range:
        raise click.BadOptionUsage(
            "-f",
            "revision-range has to be specified when -f/--force is used",
        )

    source_git_path = pathlib.Path(source_git).resolve()
    dist_git_path = pathlib.Path(dist_git).resolve()
    builder = LocalProjectBuilder(offline=True)
    api = PackitAPI(
        config=config,
        package_config=package_config,
        upstream_local_project=builder.build(
            working_dir=source_git_path,
            git_repo=CALCULATE,
        ),
        downstream_local_project=builder.build(
            working_dir=dist_git_path,
            git_repo=CALCULATE,
        ),
    )

    api.update_source_git(
        revision_range=revision_range,
        check_sync_status=not force,
    )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from pathlib import Path
from textwrap import dedent

import pytest

from packit.api import PackitAPI
from packit.exceptions import PackitConfigException
from packit.utils.commands import cwd


@pytest.mark.parametrize(
    "raw_package_config,valid, expected_output",
    [
        (
            dedent(
                """\
                dist_git_base_url: https://packit.dev/
                downstream_package_name: packit
                upstream_ref: last_commit
                upstream_package_name: packit_upstream
                allowed_gpg_keys: [gpg]
                dist_git_namespace: awesome
                notifications:
                    pull_request:
                        successful_build: true
                jobs:
                - job: copr_build
                  trigger: pull_request
                  targets: fedora-stable
            """,
            ),
            True,
            "packit.yaml is valid and ready to be used",
        ),
        (
            dedent(
                """\
                dist_git_base_url: https://packit.dev/
                downstream_package_name: packit
                upstream_ref: last_commit
                upstream_package_name: packit_upstream
                allowed_gpg_keys: [gpg]
                dist_git_namespace: awesome
                notifications:
                    pull_request:
                        successful_build: 55
            """,
            ),
            False,
            "value at index successful_build: Not a valid boolean.",
        ),
        ("", True, "packit.yaml is valid and ready to be used"),
        (
            dedent(
                """\
                dist_git_base_url: https://packit.dev/
                downstream_package_name: packit
                upstream_ref: last_commit
                upstream_package_name: packit_upstream
                allowed_gpg_keys: [gpg]
                dist_git_namespace: awesome
                synced_files:
                - a.md
                - b.md
                - c.txt
                jobs:
                - job: propose_downstream
                  trigger: release
                  dist_git_branches: fedora-latest
            """,
            ),
            True,
            "packit.yaml is valid and ready to be used",
        ),
        (
            dedent(
                """\
                dist_git_base_url: https://packit.dev/
                downstream_package_name: packit
                upstream_ref: last_commit
                upstream_package_name: packit_upstream
                allowed_gpg_keys: [gpg]
                dist_git_namespace: awesome
                synced_files:
                - src: 55
                  dest: a.md
                - b.md
                - c.txt
            """,
            ),
            False,
            "Expected 'list[str]' or 'str', got <class 'int'>.",
        ),
        (
            dedent(
                """\
                dist_git_base_url: https://packit.dev/
                downstream_package_name: packit
                upstream_ref: last_commit
                upstream_package_name: packit_upstream
                allowed_gpg_keys: [gpg]
                dist_git_namespace: awesome
                synced_files:
                - a.md
                - b.md
                - src: c.txt
                  dest: True
            """,
            ),
            False,
            "dest: Not a valid string.",
        ),
        (
            dedent(
                """\
                dist_git_base_url: https://packit.dev/
                downstream_package_name: packit
                upstream_ref: last_commit
                upstream_package_name: packit_upstream
                allowed_gpg_keys: [gpg]
                dist_git_namespace: awesome
                jobs:
                - job: tests
                  trigger: pull_request
                  targets:
                  - fedora-stable
                  - centos-stream-9
            """,
            ),
            True,
            "packit.yaml is valid and ready to be used",
        ),
        (
            dedent(
                """\
                dist_git_base_url: https://packit.dev/
                downstream_package_name: 23
                upstream_ref: last_commit
                upstream_package_name: packit_upstream
                allowed_gpg_keys: [gpg]
                dist_git_namespace: awesome
            """,
            ),
            False,
            "downstream_package_name: Not a valid string.",
        ),
        (
            dedent(
                """\
                dist_git_base_url: https://packit.dev/
                downstream_package_name: packit
                upstream_ref: last_commit
                upstream_package_name: packit_upstream
                allowed_gpg_keys: [gpg]
                dist_git_namespace: awesome
                create_pr: ""
            """,
            ),
            False,
            "create_pr: Not a valid boolean.",
        ),
        (
            dedent(
                """\
                dist_git_base_url: https://packit.dev/
                downstream_package_name: packit
                upstream_ref: last_commit
                upstream_package_name: packit_upstream
                allowed_gpg_keys: [gpg]
                dist_git_namespace: awesome
                jobs:
                - job: bodhi_update
                  trigger: commit
                  dist_git_branches:
                  - f35
                  - f36
            """,
            ),
            True,
            "packit.yaml is valid and ready to be used",
        ),
        (
            dedent(
                """\
                downstream_package_name: packit
                upstream_package_name: packit_upstream
                jobs:
                - job: propose_downstream
                  trigger: release
                  dist_git_branches:
                    fedora-rawhide:
                      fast_forward_merge_into: [fedora-all]
            """,
            ),
            True,
            "packit.yaml is valid and ready to be used",
        ),
        (
            dedent(
                """\
                downstream_package_name: packit
                upstream_package_name: packit_upstream
                jobs:
                - job: propose_downstream
                  trigger: release
                  dist_git_branches:
                    fedora-rawhide:
                      fast_forward_merge_into: [f40]
                    f40:
                      fast_forward_merge_into: [f39]
            """,
            ),
            True,
            "packit.yaml is valid and ready to be used",
        ),
        (
            dedent(
                """\
                downstream_package_name: packit
                upstream_package_name: packit_upstream
                jobs:
                - job: propose_downstream
                  trigger: release
                  dist_git_branches:
                    fedora-rawhide:
                      f40
            """,
            ),
            False,
            "Expected 'list[str]' or 'dict[str, dict[str, list]]'",
        ),
        (
            dedent(
                """\
                downstream_package_name: packit
                upstream_package_name: packit_upstream
                jobs:
                - job: propose_downstream
                  trigger: release
                  dist_git_branches:
                    fedora-rawhide:
                      open_pull_request: f40
            """,
            ),
            False,
            'Example -> "dist_git_branches":'
            ' {"fedora-rawhide": {"fast_forward_merge_into": ["f40"]}, epel9: {}}}',
        ),
        (
            dedent(
                """\
                dist_git_base_url: https://packit.dev/
                downstream_package_name: packit
                upstream_ref: last_commit
                upstream_package_name: packit_upstream
                allowed_gpg_keys: gpg
                dist_git_namespace: awesome
            """,
            ),
            False,
            "allowed_gpg_keys: Not a valid list.",
        ),
        (
            dedent(
                """\
                dist_git_base_url: https://packit.dev/
                downstream_package_name: packit
                upstream_ref: last_commit
                upstream_package_name: packit/upstream
                allowed_gpg_keys: [gpg]
                dist_git_namespace: awesome
            """,
            ),
            False,
            " Repository name must be a valid filename.",
        ),
    ],
    ids=[
        "valid_1",
        "notif_succ_build",
        "empty",
        "valid_2",
        "synced_files_src",
        "synced_files_dest",
        "valid_3",
        "downstream_name",
        "create_pr",
        "valid_4",
        "one_mapping_dist_git_prs",
        "multiple_mappings_dist_git_prs",
        "missing_fast_forward_merge_into_key",
        "wrong_fast_forward_merge_into_key",
        "allowed_gpg",
        "slash_in_package_name",
    ],
)
def test_schema_validation(tmpdir, raw_package_config, valid, expected_output):
    with cwd(tmpdir):
        Path("test_dir").mkdir()
        Path("test_dir/packit.yaml").write_text(raw_package_config)
        Path("test_dir/packit.spec").write_text("hello")
        Path("test_dir/a.md").write_text("a")
        Path("test_dir/b.md").write_text("b")
        Path("test_dir/c.txt").write_text("c")

        if valid:
            output = PackitAPI.validate_package_config(Path("test_dir"))
            assert expected_output in output
        else:
            with pytest.raises(PackitConfigException) as exc_info:
                PackitAPI.validate_package_config(Path("test_dir"))
            assert expected_output in str(exc_info.value)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT
from typing import Optional


class LabelRequirementsConfig:
    """Configuration of requirements on label."""

    def __init__(
        self,
        present: Optional[list] = None,
        absent: Optional[list] = None,
    ):
        self.present = present or []
        self.absent = absent or []


class RequirementsConfig:
    """Configuration of requirements."""

    def __init__(self, label: Optional[LabelRequirementsConfig] = None):
        self.label = label or LabelRequirementsConfig()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from subprocess import check_output

from bugzilla import Bugzilla
from flexmock import flexmock
from ogr.services.github.project import GithubProject
from requre.cassette import DataTypes
from requre.helpers.files import StoreFiles
from requre.helpers.simple_object import Simple
from requre.modules_decorate_all_methods import (
    record_git_module,
    record_requests_module,
    record_tempfile_module,
)
from requre.online_replacing import (
    apply_decorator_to_all_methods,
    replace_module_match,
)

from packit.api import PackitAPI
from tests_recording.testbase import PackitTest


@record_tempfile_module
@record_requests_module
@record_git_module
@apply_decorator_to_all_methods(
    replace_module_match(
        what="packit.utils.run_command_remote",
        decorate=Simple.decorator_plain(),
    ),
)
@apply_decorator_to_all_methods(
    replace_module_match(
        what="packit.pkgtool.PkgTool.clone",
        decorate=StoreFiles.where_arg_references(
            key_position_params_dict={"target_path": 2},
        ),
    ),
)
@apply_decorator_to_all_methods(
    replace_module_match(
        what="copr.v3.helpers.config_from_file",
        decorate=Simple.decorator_plain(),
    ),
)
class ProposeUpdate(PackitTest):
    def cassette_setup(self, cassette):
        cassette.data_miner.data_type = DataTypes.Dict

    def setUp(self):
        super().setUp()
        self.configure_git()
        self._api = None

    @property
    def api(self):
        if not self._api:
            self._api = PackitAPI(
                config=self.config,
                package_config=self.pc,
                upstream_local_project=self.lp,
            )
            self.api._up = self.upstream
            self.api._dg = self.dg
        return self._api

    def check_version_increase(self):
        """Bump version in specfile and tag it with the new version.
        Might fail if such tag already exists in requre repo.
        In that case you probably need to bump Version in fedora/python-requre.spec
        """
        filedata = self.project_specfile_location.read_text()
        # Patch the specfile with new version
        version_increase = "0.0.0"
        for line in filedata.splitlines():
            if "Version:" in line:
                version = line.rsplit(" ", 1)[1]
                v1, v2, v3 = version.split(".")
                version_increase = ".".join([v1, str(int(v2) + 1), v3])
                filedata = filedata.replace(version, version_increase)
                break
        self.project_specfile_location.write_text(filedata)
        check_output(
            f"cd {self.lp.working_dir};"
            f"git commit -m 'test change' {self._project_specfile_path.name};"
            f"git tag -a {version_increase} -m 'my version {version_increase}'",
            shell=True,
        )
        self.api.sync_release(versions=["0.8.1"], force=True)

    # We don't run this test because we haven't been able to regenerate the data for it
    # and we're not sure what it's supposed to actually test.
    # https://github.com/packit/packit/issues/1726
    def comment_in_spec(self):
        """
        change specfile a bit to have there some change, do not increase version
        """
        with self.project_specfile_location.open("a") as myfile:
            myfile.write("\n# comment\n")
        version_increase = "10.0.0"
        check_output(
            f"cd {self.lp.working_dir};"
            f"git commit -m 'test change' {self._project_specfile_path};"
            f"git tag -a {version_increase} -m 'my version {version_increase}'",
            shell=True,
        )
        self.api.sync_release(versions=["0.8.1"])

    def test_changelog_sync(self):
        """
        Bump version two times and see if the changelog is synced
        when it's configured to do so.
        """
        original_upstream_spec_content = self.api.up.absolute_specfile_path.read_text()
        check_output(
            f"cd {self.lp.working_dir};"
            f"rpmdev-bumpspec {self._project_specfile_path};"
            f"rpmdev-bumpspec {self._project_specfile_path};"
            f"git commit --no-verify -m 'test change' {self._project_specfile_path}",
            shell=True,
        )
        changed_upstream_spec_content = self.api.up.absolute_specfile_path.read_text()
        assert original_upstream_spec_content != changed_upstream_spec_content
        # mock the release as there is no real release 0
        flexmock(GithubProject).should_receive("get_release").and_return(
            flexmock(url="url"),
        )
        # not able to record the bugzilla connection, therefore mocking
        flexmock(Bugzilla).should_receive("__init__")
        flexmock(Bugzilla).should_receive("query").and_return([])

        self.api.package_config.sync_changelog = True
        self.api.sync_release(versions=["1.0.0"], use_local_content=True)
        new_downstream_spec_content = self.api.dg.absolute_specfile_path.read_text()
        assert changed_upstream_spec_content == new_downstream_spec_content



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
packit.utils.extensions
=======================

Keeps functions that extend Python collections (list to dict, get on nested dict)
and force strict evaluation (`assert_existence`).
"""

from typing import Any

from packit.exceptions import PackitException


def assert_existence(obj, name):
    """
    Force the lazy object to be evaluated.
    """
    if obj is None:
        raise PackitException(f"Object ({name}) needs to have a value.")


def nested_get(d: dict, *keys, default=None) -> Any:
    """
    recursively obtain value from nested dict

    :param d: dict
    :param keys: path within the structure
    :param default: a value to return by default

    :return: value or None
    """
    response = d
    try:
        for k in keys:
            response = response[k]
    except (KeyError, AttributeError, TypeError):
        # logger.debug("can't obtain %s: %s", k, ex)
        return default
    return response



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from pathlib import Path

import pytest
import yaml
from flexmock import flexmock

from packit.constants import DISTRO_DIR, SRC_GIT_CONFIG
from packit.exceptions import PackitException
from packit.pkgtool import PkgTool
from packit.source_git import SourceGitGenerator
from tests.integration.conftest import HELLO_RELEASE


def test_upstream_ref_not_at_head(hello_source_git_repo, hello_dist_git_repo):
    """Initializing a source-git repo fails, if the upstream_ref which
    should be the starting point of the downstream changes is not at
    the HEAD of the source-git repo.
    """
    readme = Path(hello_source_git_repo.working_dir, "README.md")
    with open(readme, "a") as fp:
        fp.write("\nSome new thing\n")
    hello_source_git_repo.git.add("README.md")
    hello_source_git_repo.git.commit(message="Add a new thing")

    sgg = SourceGitGenerator(
        config=flexmock(),
        source_git=hello_source_git_repo,
        dist_git=hello_dist_git_repo,
        upstream_ref=HELLO_RELEASE,
    )
    with pytest.raises(PackitException) as ex:
        sgg.create_from_upstream()
    assert "not pointing to the current HEAD" in str(ex.value)


def test_not_using_autosetup(hello_source_git_repo, hello_dist_git_repo):
    """Initializing a source-git repo for packages which don't use
    %autosetup in their specfile is currently not supported.
    """
    spec = Path(hello_dist_git_repo.working_dir, "hello.spec")
    content = spec.read_text().replace("%autosetup", "%setup")
    spec.write_text(content)
    hello_dist_git_repo.git.add("hello.spec")
    hello_dist_git_repo.git.commit(message="Use %setup")

    sgg = SourceGitGenerator(
        config=flexmock(),
        source_git=hello_source_git_repo,
        dist_git=hello_dist_git_repo,
        upstream_ref=HELLO_RELEASE,
    )
    with pytest.raises(PackitException) as ex:
        sgg.create_from_upstream()
    assert "not using %autosetup" in str(ex.value)


def modify_file(path: Path):
    content = path.read_text() + "\n# some change\n"
    path.write_text(content)


def create_file(path: Path):
    Path(path.parent, "new.file").touch()


@pytest.mark.parametrize("change", [modify_file, create_file])
def test_dist_git_not_pristine(hello_source_git_repo, hello_dist_git_repo, change):
    """Initialization fails if the dist-git working directory is not
    pristine, has changes and/or untracked files, in order to avoid
    tracking files in source-git which are not part of the dist-git repo
    and history.
    """
    change(Path(hello_dist_git_repo.working_dir, "hello.spec"))

    sgg = SourceGitGenerator(
        config=flexmock(),
        source_git=hello_source_git_repo,
        dist_git=hello_dist_git_repo,
        upstream_ref=HELLO_RELEASE,
        pkg_name="hello",
    )
    with pytest.raises(PackitException) as ex:
        sgg.create_from_upstream()
    assert "is not pristine" in str(ex.value)


def download_sources(from_repo, to_repo):
    def funk():
        from_repo.git.archive(
            HELLO_RELEASE,
            format="tar.gz",
            prefix=f"hello-{HELLO_RELEASE}/",
            output=f"{to_repo.working_dir}/hello-{HELLO_RELEASE}.tar.gz",
        )

    return funk


def check_source_git_config(source_git_config):
    assert source_git_config["upstream_project_url"] == "https://example.com/hello.git"
    assert source_git_config["upstream_ref"] == HELLO_RELEASE
    assert source_git_config["downstream_package_name"] == "hello"
    assert source_git_config["specfile_path"] == ".distro/hello.spec"
    assert source_git_config["patch_generation_ignore_paths"] == [DISTRO_DIR]
    assert source_git_config["sync_changelog"] is True
    assert source_git_config["files_to_sync"] == [
        {
            "src": ".distro/",
            "dest": ".",
            "delete": True,
            "filters": [
                "protect .git*",
                "protect sources",
                f"exclude {SRC_GIT_CONFIG}",
                "exclude .gitignore",
            ],
        },
    ]
    assert source_git_config["sources"][0]["path"] == f"hello-{HELLO_RELEASE}.tar.gz"


@pytest.mark.skipif(
    # on rawhide the version can contain letters:
    # >>> yaml.__version__
    # '6.0b1'
    list(map(int, yaml.__version__[:3].split("."))) < [5, 1],
    reason="Requires PyYAML 5.1 or higher.",
)
def test_create_from_upstream_no_patch(hello_source_git_repo, hello_dist_git_repo):
    """A source-git repo is properly initialized from a dist-git repo.
    - No downstream patches.
    """
    spec = Path(hello_dist_git_repo.working_dir, "hello.spec")
    content = [
        line for line in spec.read_text().splitlines() if not line.startswith("Patch")
    ]
    spec.write_text("\n".join(content) + "\n")
    Path(hello_dist_git_repo.working_dir, "turn-into-fedora.patch").unlink()
    hello_dist_git_repo.git.add(".")
    hello_dist_git_repo.git.commit(message="Remove the patch")

    flexmock(
        PkgTool,
        sources=download_sources(hello_source_git_repo, hello_dist_git_repo),
    )
    sgg = SourceGitGenerator(
        config=flexmock(fas_user=None, pkg_tool="fedpkg"),
        source_git=hello_source_git_repo,
        dist_git=hello_dist_git_repo,
        upstream_ref=HELLO_RELEASE,
        pkg_name="hello",
    )
    sgg.create_from_upstream()
    source_git_config = yaml.safe_load(
        Path(hello_source_git_repo.working_dir, DISTRO_DIR, SRC_GIT_CONFIG).read_text(),
    )
    check_source_git_config(source_git_config)
    assert source_git_config["patch_generation_patch_id_digits"] == 1
    assert (
        f"\nFrom-dist-git-commit: {hello_dist_git_repo.head.commit.hexsha}\n"
        in hello_source_git_repo.head.commit.message
    )


@pytest.mark.skipif(
    # on rawhide the version can contain letters:
    list(map(int, yaml.__version__[:3].split("."))) < [5, 1],
    reason="Requires PyYAML 5.1 or higher.",
)
def test_create_from_upstream_with_patch(hello_source_git_repo, hello_dist_git_repo):
    """A source-git repo is properly initialized from a dist-git repo.
    - A few downstream patches.
    """
    flexmock(
        PkgTool,
        sources=download_sources(hello_source_git_repo, hello_dist_git_repo),
    )
    sgg = SourceGitGenerator(
        config=flexmock(fas_user=None, pkg_tool="fedpkg"),
        source_git=hello_source_git_repo,
        dist_git=hello_dist_git_repo,
        upstream_ref=HELLO_RELEASE,
        pkg_name="hello",
    )
    sgg.create_from_upstream()
    source_git_config = yaml.safe_load(
        Path(hello_source_git_repo.working_dir, DISTRO_DIR, SRC_GIT_CONFIG).read_text(),
    )
    check_source_git_config(source_git_config)
    assert source_git_config["patch_generation_patch_id_digits"] == 4

    assert (
        Path(hello_source_git_repo.working_dir, DISTRO_DIR, ".gitignore").read_text()
        == """\
# Reset gitignore rules
!*
"""
    )
    assert not Path(hello_source_git_repo.working_dir, DISTRO_DIR, "sources").exists()
    assert not Path(hello_source_git_repo.working_dir, DISTRO_DIR, ".git").exists()

    assert (
        "Hello Fedora Linux"
        in Path(hello_source_git_repo.working_dir, "hello.rs").read_text()
    )

    commit_messsage_lines = hello_source_git_repo.commit("HEAD~2").message.splitlines()
    assert "Patch-name: turn-into-fedora.patch" in commit_messsage_lines
    assert "Patch-id: 1" in commit_messsage_lines
    assert "Patch-status: |" in commit_messsage_lines
    assert (
        f"From-dist-git-commit: {hello_dist_git_repo.head.commit.hexsha}"
        in commit_messsage_lines
    )

    commit_messsage_lines = hello_source_git_repo.commit("HEAD~1").message.splitlines()
    assert "Patch-name: from-git.patch" in commit_messsage_lines
    assert "Patch-id: 2" in commit_messsage_lines
    assert "Patch-status: |" in commit_messsage_lines
    assert (
        f"From-dist-git-commit: {hello_dist_git_repo.head.commit.hexsha}"
        in commit_messsage_lines
    )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from pathlib import Path
from typing import Optional, Union

from flexmock import flexmock

from packit.local_project import LocalProject
from packit.utils import commands


def test_basic_build(
    cwd_upstream_or_distgit,
    api_instance,
    mock_remote_functionality_upstream,
):
    u, d, api = api_instance
    flexmock(api).should_receive("init_kerberos_ticket").at_least().once()
    flexmock(commands).should_receive("run_command_remote").with_args(
        cmd=["fedpkg", "build", "--scratch", "--nowait", "--target", "asdqwe"],
        cwd=api.dg.local_project.working_dir,
        error_message="Submission of build to koji failed.",
        fail=True,
        output=True,
        print_live=True,
    ).once().and_return(flexmock(stdout=""))

    api.build("main", scratch=True, nowait=True, koji_target="asdqwe")


def test_build_from_upstream(
    cwd_upstream_or_distgit,
    api_instance,
    mock_remote_functionality_upstream,
):
    u, d, api = api_instance

    def mocked_run_command(
        cmd: Union[list[str], str],
        error_message: Optional[str] = None,
        cwd: Optional[Union[str, Path]] = None,
        fail: bool = True,
        output: bool = False,
        env: Optional[dict] = None,
        decode=True,
        print_live=False,
    ):
        assert cmd[:-1] == ["koji", "build", "--scratch", "--nowait", "hadron-collider"]
        srpm_path = cmd[-1]
        assert Path(srpm_path).is_file()
        assert srpm_path.endswith(".src.rpm")
        assert cwd == api.up.local_project.working_dir
        return flexmock(
            success=True,
            stdout="\n\nLink to koji build: https://koji...\n",
        )

    flexmock(commands, run_command_remote=mocked_run_command)
    flexmock(LocalProject).should_receive("free_resources")
    api.build(
        "master",
        scratch=True,
        nowait=True,
        from_upstream=True,
        koji_target="hadron-collider",
    )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT
from typing import Optional


class PullRequestNotificationsConfig:
    """Configuration of commenting on pull requests."""

    def __init__(self, successful_build: bool = False):
        self.successful_build = successful_build


class FailureCommentNotificationsConfig:
    """Configuration of the failure comment."""

    def __init__(self, message: Optional[str] = None):
        self.message = message


class FailureIssueNotificationsConfig:
    """Configuration of the failure issue for upstream."""

    def __init__(self, create: bool = True):
        self.create = create


class NotificationsConfig:
    """Configuration of notifications."""

    def __init__(
        self,
        pull_request: Optional[PullRequestNotificationsConfig] = None,
        failure_comment: Optional[FailureCommentNotificationsConfig] = None,
        failure_issue: Optional[FailureIssueNotificationsConfig] = None,
    ):
        self.pull_request = pull_request or PullRequestNotificationsConfig()
        self.failure_comment = failure_comment or FailureCommentNotificationsConfig()
        self.failure_issue = failure_issue or FailureIssueNotificationsConfig()
